"""ASAM ODS Protocol Buffer interfaces

This module provides convenient access to ASAM ODS protobuf definitions
with lazy loading for better performance.

Example::

    from odsbox.proto import ods, ods_security

    # Create a model instance
    model = ods.Model()

"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from . import ods_notification_pb2 as ods_notification
    from . import ods_pb2 as ods
    from . import ods_security_pb2 as ods_security


def __getattr__(name: str):
    """Lazy import for better performance and optional dependencies"""
    if name == "ods":
        from . import ods_pb2 as ods

        return ods
    elif name == "ods_security":
        from . import ods_security_pb2 as ods_security

        return ods_security
    elif name == "ods_notification":
        from . import ods_notification_pb2 as ods_notification

        return ods_notification
    elif name == "ods_external_data":
        try:
            from . import ods_external_data_pb2 as ods_external_data

            return ods_external_data
        except ImportError as e:
            raise ImportError("ods_external_data requires grpcio. Install with: pip install odsbox[exd-data]") from e
    elif name == "ods_external_data_grpc":
        try:
            from . import ods_external_data_pb2_grpc as ods_external_data_grpc

            return ods_external_data_grpc
        except ImportError as e:
            raise ImportError(
                "ods_external_data_grpc requires grpcio. Install with: pip install odsbox[exd-data]"
            ) from e
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def __dir__():
    """Return list of available attributes for tab completion"""
    return ["ods", "ods_security", "ods_notification", "ods_external_data", "ods_external_data_grpc"]


# Define what gets imported with "from odsbox.proto import *"
__all__ = ["ods", "ods_security", "ods_notification", "ods_external_data", "ods_external_data_grpc"]



"""odsbox - Toolbox for accessing ASAM ODS servers using the HTTP API

This package provides convenient access to ASAM ODS servers with lazy loading
for better performance.

Example::

    from odsbox import ConI

    with ConI(url="http://localhost:8087/api", auth=("sa", "sa")) as con_i:
        units = con_i.query_data({"AoUnit": {}})

"""

from __future__ import annotations

from typing import TYPE_CHECKING

__version__ = "1.0.14"""

if TYPE_CHECKING:
    from .con_i import ConI


def __getattr__(name: str):
    """Lazy import for better performance"""
    if name == "ConI":
        from .con_i import ConI

        return ConI
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def __dir__():
    """Return list of available attributes for tab completion"""
    return ["ConI", "__version__"]


# Define what gets imported with "from odsbox import *"
__all__ = ["ConI"]



"""Integration test for ASAM ODS session"""

from __future__ import annotations

import pytest

from odsbox.con_i import ConI
from odsbox.unit_catalog import UnitCatalog


def __create_con_i():
    """Create a connection session for an ASAM ODS server"""
    return ConI("https://docker.peak-solution.de:10032/api", ("Demo", "mdm"))


def test_unit_get():
    with __create_con_i() as con_i:
        unit_catalog = UnitCatalog(con_i)

        assert unit_catalog.get("gradC") is not None
        assert unit_catalog.get("s") is not None
        assert unit_catalog.get("1/min") is not None
        assert unit_catalog.get("V") is not None
        assert unit_catalog.get("J") is not None
        assert unit_catalog.get("A") is not None
        assert unit_catalog.get("UnitDoesNotExit") is None
        assert unit_catalog.get("UnitDoesNotExit") is None


@pytest.mark.integration
def test_unit_get_or_create():
    with __create_con_i() as con_i:
        unit_catalog = UnitCatalog(con_i)

        assert unit_catalog.get_or_create("°C") > 0
        assert unit_catalog.get_or_create("s") > 0
        assert unit_catalog.get_or_create("rpm") > 0
        assert unit_catalog.get_or_create("V") > 0
        assert unit_catalog.get_or_create("J") > 0
        assert unit_catalog.get_or_create("A") > 0



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class SecurityTypeEnum(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    ST_ACL: _ClassVar[SecurityTypeEnum]
    ST_TPLI: _ClassVar[SecurityTypeEnum]
    ST_EFF: _ClassVar[SecurityTypeEnum]

class DataObjectTypeEnum(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    DOT_APPLICATION_ELEMENT: _ClassVar[DataObjectTypeEnum]
    DOT_APPLICATION_ATTRIBUTE: _ClassVar[DataObjectTypeEnum]
    DOT_INSTANCE: _ClassVar[DataObjectTypeEnum]
ST_ACL: SecurityTypeEnum
ST_TPLI: SecurityTypeEnum
ST_EFF: SecurityTypeEnum
DOT_APPLICATION_ELEMENT: DataObjectTypeEnum
DOT_APPLICATION_ATTRIBUTE: DataObjectTypeEnum
DOT_INSTANCE: DataObjectTypeEnum

class SecurityReadRequest(_message.Message):
    __slots__ = ("security_request_type", "data_object_type", "application_element", "application_attribute", "instance", "user_group_id")
    class DataObjectApplicationElement(_message.Message):
        __slots__ = ("aid",)
        AID_FIELD_NUMBER: _ClassVar[int]
        aid: int
        def __init__(self, aid: _Optional[int] = ...) -> None: ...
    class DataObjectApplicationAttribute(_message.Message):
        __slots__ = ("aid", "attribute_names")
        AID_FIELD_NUMBER: _ClassVar[int]
        ATTRIBUTE_NAMES_FIELD_NUMBER: _ClassVar[int]
        aid: int
        attribute_names: _containers.RepeatedScalarFieldContainer[str]
        def __init__(self, aid: _Optional[int] = ..., attribute_names: _Optional[_Iterable[str]] = ...) -> None: ...
    class DataObjectInstance(_message.Message):
        __slots__ = ("aid", "iids")
        AID_FIELD_NUMBER: _ClassVar[int]
        IIDS_FIELD_NUMBER: _ClassVar[int]
        aid: int
        iids: _containers.RepeatedScalarFieldContainer[int]
        def __init__(self, aid: _Optional[int] = ..., iids: _Optional[_Iterable[int]] = ...) -> None: ...
    SECURITY_REQUEST_TYPE_FIELD_NUMBER: _ClassVar[int]
    DATA_OBJECT_TYPE_FIELD_NUMBER: _ClassVar[int]
    APPLICATION_ELEMENT_FIELD_NUMBER: _ClassVar[int]
    APPLICATION_ATTRIBUTE_FIELD_NUMBER: _ClassVar[int]
    INSTANCE_FIELD_NUMBER: _ClassVar[int]
    USER_GROUP_ID_FIELD_NUMBER: _ClassVar[int]
    security_request_type: SecurityTypeEnum
    data_object_type: DataObjectTypeEnum
    application_element: SecurityReadRequest.DataObjectApplicationElement
    application_attribute: SecurityReadRequest.DataObjectApplicationAttribute
    instance: SecurityReadRequest.DataObjectInstance
    user_group_id: int
    def __init__(self, security_request_type: _Optional[_Union[SecurityTypeEnum, str]] = ..., data_object_type: _Optional[_Union[DataObjectTypeEnum, str]] = ..., application_element: _Optional[_Union[SecurityReadRequest.DataObjectApplicationElement, _Mapping]] = ..., application_attribute: _Optional[_Union[SecurityReadRequest.DataObjectApplicationAttribute, _Mapping]] = ..., instance: _Optional[_Union[SecurityReadRequest.DataObjectInstance, _Mapping]] = ..., user_group_id: _Optional[int] = ...) -> None: ...

class SecurityEntry(_message.Message):
    __slots__ = ("aid", "aa_name", "iid", "usergroup_id", "rights", "ref_aid")
    AID_FIELD_NUMBER: _ClassVar[int]
    AA_NAME_FIELD_NUMBER: _ClassVar[int]
    IID_FIELD_NUMBER: _ClassVar[int]
    USERGROUP_ID_FIELD_NUMBER: _ClassVar[int]
    RIGHTS_FIELD_NUMBER: _ClassVar[int]
    REF_AID_FIELD_NUMBER: _ClassVar[int]
    aid: int
    aa_name: str
    iid: int
    usergroup_id: int
    rights: int
    ref_aid: int
    def __init__(self, aid: _Optional[int] = ..., aa_name: _Optional[str] = ..., iid: _Optional[int] = ..., usergroup_id: _Optional[int] = ..., rights: _Optional[int] = ..., ref_aid: _Optional[int] = ...) -> None: ...

class SecurityInfo(_message.Message):
    __slots__ = ("security_type", "data_object_type", "entries")
    SECURITY_TYPE_FIELD_NUMBER: _ClassVar[int]
    DATA_OBJECT_TYPE_FIELD_NUMBER: _ClassVar[int]
    ENTRIES_FIELD_NUMBER: _ClassVar[int]
    security_type: SecurityTypeEnum
    data_object_type: DataObjectTypeEnum
    entries: _containers.RepeatedCompositeFieldContainer[SecurityEntry]
    def __init__(self, security_type: _Optional[_Union[SecurityTypeEnum, str]] = ..., data_object_type: _Optional[_Union[DataObjectTypeEnum, str]] = ..., entries: _Optional[_Iterable[_Union[SecurityEntry, _Mapping]]] = ...) -> None: ...

class SecurityWriteRequest(_message.Message):
    __slots__ = ("security_request_type", "data_object_type", "entries", "modify_type")
    class SecurityModifyTypeEnum(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        SST_ADD: _ClassVar[SecurityWriteRequest.SecurityModifyTypeEnum]
        SST_REMOVE: _ClassVar[SecurityWriteRequest.SecurityModifyTypeEnum]
        SST_SET: _ClassVar[SecurityWriteRequest.SecurityModifyTypeEnum]
    SST_ADD: SecurityWriteRequest.SecurityModifyTypeEnum
    SST_REMOVE: SecurityWriteRequest.SecurityModifyTypeEnum
    SST_SET: SecurityWriteRequest.SecurityModifyTypeEnum
    SECURITY_REQUEST_TYPE_FIELD_NUMBER: _ClassVar[int]
    DATA_OBJECT_TYPE_FIELD_NUMBER: _ClassVar[int]
    ENTRIES_FIELD_NUMBER: _ClassVar[int]
    MODIFY_TYPE_FIELD_NUMBER: _ClassVar[int]
    security_request_type: SecurityTypeEnum
    data_object_type: DataObjectTypeEnum
    entries: _containers.RepeatedCompositeFieldContainer[SecurityEntry]
    modify_type: SecurityWriteRequest.SecurityModifyTypeEnum
    def __init__(self, security_request_type: _Optional[_Union[SecurityTypeEnum, str]] = ..., data_object_type: _Optional[_Union[DataObjectTypeEnum, str]] = ..., entries: _Optional[_Iterable[_Union[SecurityEntry, _Mapping]]] = ..., modify_type: _Optional[_Union[SecurityWriteRequest.SecurityModifyTypeEnum, str]] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: odsbox/proto/ods_notification.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n#odsbox/proto/ods_notification.proto\x12\x10ods.notification\"\xa8\x02\n\x13RegistrationRequest\x12H\n\x04mode\x18\x01 \x01(\x0e\x32:.ods.notification.RegistrationRequest.NotificationModeEnum\x12\x42\n\x07\x65ntries\x18\x02 \x03(\x0b\x32\x31.ods.notification.RegistrationRequest.TypeElement\x1aQ\n\x0bTypeElement\x12\x34\n\x04type\x18\x01 \x01(\x0e\x32&.ods.notification.NotificationTypeEnum\x12\x0c\n\x04\x61ids\x18\x02 \x03(\x03\"0\n\x14NotificationModeEnum\x12\x0b\n\x07NM_POOL\x10\x00\x12\x0b\n\x07NM_PUSH\x10\x01\"\x91\x01\n\x0cNotification\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x34\n\x04type\x18\x02 \x01(\x0e\x32&.ods.notification.NotificationTypeEnum\x12\x0b\n\x03\x61id\x18\x03 \x01(\x03\x12\x0c\n\x04iids\x18\x04 \x03(\x03\x12\x11\n\ttimestamp\x18\x05 \x01(\t\x12\x0f\n\x07user_id\x18\x06 \x01(\x03\"B\n\x10NotificationPool\x12.\n\x06\x65vents\x18\x01 \x03(\x0b\x32\x1e.ods.notification.Notification*k\n\x14NotificationTypeEnum\x12\n\n\x06NT_ALL\x10\x00\x12\n\n\x06NT_NEW\x10\x01\x12\r\n\tNT_MODIFY\x10\x02\x12\r\n\tNT_DELETE\x10\x03\x12\x0c\n\x08NT_MODEL\x10\x04\x12\x0f\n\x0bNT_SECURITY\x10\x05\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'odsbox.proto.ods_notification_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_NOTIFICATIONTYPEENUM']._serialized_start=572
  _globals['_NOTIFICATIONTYPEENUM']._serialized_end=679
  _globals['_REGISTRATIONREQUEST']._serialized_start=58
  _globals['_REGISTRATIONREQUEST']._serialized_end=354
  _globals['_REGISTRATIONREQUEST_TYPEELEMENT']._serialized_start=223
  _globals['_REGISTRATIONREQUEST_TYPEELEMENT']._serialized_end=304
  _globals['_REGISTRATIONREQUEST_NOTIFICATIONMODEENUM']._serialized_start=306
  _globals['_REGISTRATIONREQUEST_NOTIFICATIONMODEENUM']._serialized_end=354
  _globals['_NOTIFICATION']._serialized_start=357
  _globals['_NOTIFICATION']._serialized_end=502
  _globals['_NOTIFICATIONPOOL']._serialized_start=504
  _globals['_NOTIFICATIONPOOL']._serialized_end=570
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import os
from pathlib import Path

from google.protobuf.json_format import Parse

from odsbox.datamatrices_to_pandas import to_pandas
from odsbox.model_cache import ModelCache
from odsbox.proto.ods_pb2 import DataMatrices, Model


def __get_model(model_file_name):
    model_file = os.path.join(os.path.abspath(os.path.dirname(__file__)), "test_data", model_file_name)
    model = Model()
    Parse(Path(model_file).read_text(encoding="utf-8"), model)
    return model


def test_map_enum_values():
    mc = ModelCache(__get_model("application_model.json"))

    mq_entity = mc.entity_by_base_name("AoMeasurementQuantity")
    assert mq_entity is not None
    datatype_attribute = mc.attribute_by_base_name(mq_entity, "datatype")
    assert datatype_attribute is not None

    dms = DataMatrices()
    dm = dms.matrices.add(aid=mq_entity.aid, name=mq_entity.name, base_name=mq_entity.base_name)
    dm.columns.add(
        name=datatype_attribute.name, base_name=datatype_attribute.base_name, data_type=datatype_attribute.data_type
    ).long_array.values[:] = [10, 2, 1, 30, 7, 0, 8, 9, 5, 28, 12, 4, 6, 13, 3, 14, 11]

    assert to_pandas(dms).to_dict() == {
        "MeaQuantity.DataType": {
            0: 10,
            1: 2,
            2: 1,
            3: 30,
            4: 7,
            5: 0,
            6: 8,
            7: 9,
            8: 5,
            9: 28,
            10: 12,
            11: 4,
            12: 6,
            13: 13,
            14: 3,
            15: 14,
            16: 11,
        }
    }

    assert to_pandas(dms, mc, True).to_dict() == {
        "MeaQuantity.DataType": {
            0: "DT_DATE",
            1: "DT_SHORT",
            2: "DT_STRING",
            3: "DT_ENUM",
            4: "DT_DOUBLE",
            5: "DT_UNKNOWN",
            6: "DT_LONGLONG",
            7: "DT_ID",
            8: "DT_BYTE",
            9: "DT_EXTERNALREFERENCE",
            10: "DT_BLOB",
            11: "DT_BOOLEAN",
            12: "DT_LONG",
            13: "DT_COMPLEX",
            14: "DT_FLOAT",
            15: "DT_DCOMPLEX",
            16: "DT_BYTESTR",
        }
    }


def test_map_enum_values_ds():
    mc = ModelCache(__get_model("application_model.json"))

    entity = mc.entity("sound_measurement")
    assert entity is not None
    attribute_dt = mc.attribute(entity, "DT_ENUM_VALUES")
    assert attribute_dt is not None
    attribute_ds = mc.attribute(entity, "DS_ENUM_VALUES")
    assert attribute_ds is not None

    dms = DataMatrices()
    dm = dms.matrices.add(aid=entity.aid, name=entity.name, base_name=entity.base_name)
    dm.columns.add(
        name=attribute_dt.name, base_name=attribute_dt.base_name, data_type=attribute_dt.data_type
    ).long_array.values[:] = [0, 1, 2]
    values = dm.columns.add(
        name=attribute_ds.name, base_name=attribute_ds.base_name, data_type=attribute_ds.data_type
    ).long_arrays.values
    values.add().values.extend([0, 1, 2])
    values.add().values.extend([2, 1, 0])
    values.add().values.extend([2, 0, 2])

    assert to_pandas(dms).to_dict() == {
        "sound_measurement.DT_ENUM_VALUES": {0: 0, 1: 1, 2: 2},
        "sound_measurement.DS_ENUM_VALUES": {0: [0, 1, 2], 1: [2, 1, 0], 2: [2, 0, 2]},
    }

    assert to_pandas(dms, mc, True).to_dict() == {
        "sound_measurement.DT_ENUM_VALUES": {0: "editing", 1: "valid", 2: "archive"},
        "sound_measurement.DS_ENUM_VALUES": {
            0: ["editing", "valid", "archive"],
            1: ["archive", "valid", "editing"],
            2: ["archive", "editing", "archive"],
        },
    }



from __future__ import annotations

from unittest import mock

import odsbox.proto.ods_security_pb2 as ods_security
from odsbox.security import Security


class DummyResponse:
    __slots__ = ("content",)

    def __init__(self, content):
        object.__setattr__(self, "content", content)

    def __setattr__(self, key, value):
        raise AttributeError(f"{self.__class__.__name__} is immutable")


def test_security_read_calls_ods_post_request_and_parses_response():
    con_i = mock.Mock()
    security_read_request = ods_security.SecurityReadRequest(
        data_object_type=ods_security.DataObjectTypeEnum.DOT_APPLICATION_ELEMENT
    )
    expected_security_info = ods_security.SecurityInfo(
        entries=[ods_security.SecurityEntry(rights=Security.Right.READ | Security.Right.UPDATE)]
    )
    serialized = expected_security_info.SerializeToString()
    con_i.ods_post_request.return_value = DummyResponse(serialized)

    security: Security = Security(con_i)
    result: ods_security.SecurityInfo = security.security_read(security_read_request)
    print(result)

    con_i.ods_post_request.assert_called_once_with("security-read", security_read_request)
    assert isinstance(result, ods_security.SecurityInfo)
    assert Security.Right.READ in Security.Right(result.entries[0].rights)


def test_security_update_calls_ods_post_request():
    con_i = mock.Mock()
    security_write_request = ods_security.SecurityWriteRequest()
    security = Security(con_i)

    security.security_update(security_write_request)

    con_i.ods_post_request.assert_called_once_with("security-update", security_write_request)


def test_initial_rights_calls_ods_post_request():
    con_i = mock.Mock()
    security_write_request = ods_security.SecurityWriteRequest()
    security = Security(con_i)

    security.initial_rights(security_write_request)

    con_i.ods_post_request.assert_called_once_with("initial-rights", security_write_request)
